/*
 *  linux/fs/aclfs/inode.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 */

#include <linux/sched.h>
#include <linux/aclfs_fs.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/locks.h>

#include <asm/system.h>
#include <asm/segment.h>


void aclfs_put_inode(struct inode *inode)
{
	if (inode->i_nlink)
		return;
	inode->i_size = 0;
	aclfs_truncate(inode);
	aclfs_free_inode(inode);
}

void aclfs_put_super(struct super_block *sb)
{
	int i;

	lock_super(sb);
	sb->s_dev = 0;
	for(i = 0 ; i < ACLFS_I_MAP_SLOTS ; i++)
		brelse(sb->u.aclfs_sb.s_imap[i]);
	for(i = 0 ; i < ACLFS_Z_MAP_SLOTS ; i++)
		brelse(sb->u.aclfs_sb.s_zmap[i]);
	unlock_super(sb);
	return;
}

static struct super_operations aclfs_sops = { 
	aclfs_read_inode,
	aclfs_notify_change,
	aclfs_write_inode,
	aclfs_put_inode,
	aclfs_put_super,
	NULL,
	aclfs_statfs
};

struct super_block *aclfs_read_super(struct super_block *s,void *data)
{
	struct buffer_head *bh;
	struct aclfs_super_block *ms;
	int i,dev=s->s_dev,block;

	/* this was 32 for the minix stuff -- we doubled the size :) */
	if (64 != sizeof (struct aclfs_inode)) {
		printk("bad i-node size\n");
		return NULL;
	}
	lock_super(s);
	if (!(bh = bread(dev,1,BLOCK_SIZE))) {
		s->s_dev=0;
		unlock_super(s);
		printk("ACLFS-fs: unable to read superblock\n");
		return NULL;
	}
	ms = (struct aclfs_super_block *) bh->b_data;
	s->s_blocksize = 1024;
	s->u.aclfs_sb.s_ninodes = ms->s_ninodes;
	s->u.aclfs_sb.s_nzones = ms->s_nzones;
	s->u.aclfs_sb.s_imap_blocks = ms->s_imap_blocks;
	s->u.aclfs_sb.s_zmap_blocks = ms->s_zmap_blocks;
	s->u.aclfs_sb.s_firstdatazone = ms->s_firstdatazone;
	s->u.aclfs_sb.s_log_zone_size = ms->s_log_zone_size;
	s->u.aclfs_sb.s_max_size = ms->s_max_size;
	s->s_magic = ms->s_magic;
	brelse(bh);
	if (s->s_magic != ACLFS_SUPER_MAGIC) {
		s->s_dev = 0;
		unlock_super(s);
		printk("ACLFS-fs magic match failed\n");
		return NULL;
	}
	for (i=0;i < ACLFS_I_MAP_SLOTS;i++)
		s->u.aclfs_sb.s_imap[i] = NULL;
	for (i=0;i < ACLFS_Z_MAP_SLOTS;i++)
		s->u.aclfs_sb.s_zmap[i] = NULL;
	block=2;
	for (i=0 ; i < s->u.aclfs_sb.s_imap_blocks ; i++)
		if ((s->u.aclfs_sb.s_imap[i]=bread(dev,block,BLOCK_SIZE)) != NULL)
			block++;
		else
			break;
	for (i=0 ; i < s->u.aclfs_sb.s_zmap_blocks ; i++)
		if ((s->u.aclfs_sb.s_zmap[i]=bread(dev,block,BLOCK_SIZE)) != NULL)
			block++;
		else
			break;
	if (block != 2+s->u.aclfs_sb.s_imap_blocks+s->u.aclfs_sb.s_zmap_blocks) {
		for(i=0;i<ACLFS_I_MAP_SLOTS;i++)
			brelse(s->u.aclfs_sb.s_imap[i]);
		for(i=0;i<ACLFS_Z_MAP_SLOTS;i++)
			brelse(s->u.aclfs_sb.s_zmap[i]);
		s->s_dev=0;
		unlock_super(s);
		printk("ACLFS-fs: bad superblock or unable to read bitmaps\n");
		return NULL;
	}
	s->u.aclfs_sb.s_imap[0]->b_data[0] |= 1;
	s->u.aclfs_sb.s_zmap[0]->b_data[0] |= 1;
	/* set up enough so that it can read an inode */
	s->s_dev = dev;
	s->s_op = &aclfs_sops;
	s->s_mounted = iget(s,ACLFS_ROOT_INO);
	unlock_super(s);
	if (!s->s_mounted) {
		s->s_dev = 0;
		printk("ACLFS-fs: get root inode failed\n");
		return NULL;
	}
	return s;
}

void aclfs_statfs(struct super_block *sb, struct statfs *buf)
{
	long tmp;

	put_fs_long(ACLFS_SUPER_MAGIC, &buf->f_type);
	put_fs_long(1024, &buf->f_bsize);
	put_fs_long(sb->u.aclfs_sb.s_nzones << sb->u.aclfs_sb.s_log_zone_size, &buf->f_blocks);
	tmp = aclfs_count_free_blocks(sb);
	put_fs_long(tmp, &buf->f_bfree);
	put_fs_long(tmp, &buf->f_bavail);
	put_fs_long(sb->u.aclfs_sb.s_ninodes, &buf->f_files);
	put_fs_long(aclfs_count_free_inodes(sb), &buf->f_ffree);
	/* Don't know what value to put in buf->f_fsid */
}

#define inode_bmap(inode,nr) ((inode)->u.aclfs_i.i_data[(nr)])

static int block_bmap(struct buffer_head * bh, int nr)
{
	int tmp;

	if (!bh)
		return 0;
	tmp = ((unsigned short *) bh->b_data)[nr];
	brelse(bh);
	return tmp;
}

int aclfs_bmap(struct inode * inode,int block)
{
	int i;

	if (block<0) {
		printk("aclfs_bmap: block<0");
		return 0;
	}
	if (block >= 7+512+512*512) {
		printk("aclfs_bmap: block>big");
		return 0;
	}
	if (block < 7)
		return inode_bmap(inode,block);
	block -= 7;
	if (block < 512) {
		i = inode_bmap(inode,7);
		if (!i)
			return 0;
		return block_bmap(bread(inode->i_dev,i,BLOCK_SIZE),block);
	}
	block -= 512;
	i = inode_bmap(inode,8);
	if (!i)
		return 0;
	i = block_bmap(bread(inode->i_dev,i,BLOCK_SIZE),block>>9);
	if (!i)
		return 0;
	return block_bmap(bread(inode->i_dev,i,BLOCK_SIZE),block & 511);
}

static struct buffer_head * inode_getblk(struct inode * inode, int nr, int create)
{
	int tmp;
	unsigned short *p;
	struct buffer_head * result;

	p = inode->u.aclfs_i.i_data + nr;
repeat:
	tmp = *p;
	if (tmp) {
		result = getblk(inode->i_dev, tmp, BLOCK_SIZE);
		if (tmp == *p)
			return result;
		brelse(result);
		goto repeat;
	}
	if (!create)
		return NULL;
	tmp = aclfs_new_block(inode->i_sb);
	if (!tmp)
		return NULL;
	result = getblk(inode->i_dev, tmp, BLOCK_SIZE);
	if (*p) {
		aclfs_free_block(inode->i_sb,tmp);
		brelse(result);
		goto repeat;
	}
	*p = tmp;
	inode->i_ctime = CURRENT_TIME;
	inode->i_dirt = 1;
	return result;
}

static struct buffer_head * block_getblk(struct inode * inode, 
	struct buffer_head * bh, int nr, int create)
{
	int tmp;
	unsigned short *p;
	struct buffer_head * result;

	if (!bh)
		return NULL;
	if (!bh->b_uptodate) {
		ll_rw_block(READ, 1, &bh);
		wait_on_buffer(bh);
		if (!bh->b_uptodate) {
			brelse(bh);
			return NULL;
		}
	}
	p = nr + (unsigned short *) bh->b_data;
repeat:
	tmp = *p;
	if (tmp) {
		result = getblk(bh->b_dev, tmp, BLOCK_SIZE);
		if (tmp == *p) {
			brelse(bh);
			return result;
		}
		brelse(result);
		goto repeat;
	}
	if (!create) {
		brelse(bh);
		return NULL;
	}
	tmp = aclfs_new_block(inode->i_sb);
	if (!tmp) {
		brelse(bh);
		return NULL;
	}
	result = getblk(bh->b_dev, tmp, BLOCK_SIZE);
	if (*p) {
		aclfs_free_block(inode->i_sb,tmp);
		brelse(result);
		goto repeat;
	}
	*p = tmp;
	bh->b_dirt = 1;
	brelse(bh);
	return result;
}

struct buffer_head * aclfs_getblk(struct inode * inode, int block, int create)
{
	struct buffer_head * bh;

	if (block<0) {
		printk("aclfs_getblk: block<0");
		return NULL;
	}
	if (block >= 7+512+512*512) {
		printk("aclfs_getblk: block>big");
		return NULL;
	}
	if (block < 7)
		return inode_getblk(inode,block,create);
	block -= 7;
	if (block < 512) {
		bh = inode_getblk(inode,7,create);
		return block_getblk(inode, bh, block, create);
	}
	block -= 512;
	bh = inode_getblk(inode,8,create);
	bh = block_getblk(inode, bh, block>>9, create);
	return block_getblk(inode, bh, block & 511, create);
}

struct buffer_head * aclfs_bread(struct inode * inode, int block, int create)
{
	struct buffer_head * bh;

	bh = aclfs_getblk(inode,block,create);
	if (!bh || bh->b_uptodate)
		return bh;
	ll_rw_block(READ, 1, &bh);
	wait_on_buffer(bh);
	if (bh->b_uptodate)
		return bh;
	brelse(bh);
	return NULL;
}

void aclfs_read_inode(struct inode * inode)
{
	struct buffer_head * bh;
	struct aclfs_inode * raw_inode;
	int block, ino, x = 0, y;


#ifdef VERBOSE_INODE_OPS
        printk("read_inode() -- reading inode\n");
#endif

	ino = inode->i_ino;
	inode->i_op = NULL;
	inode->i_mode = 0;
	if (!ino || ino >= inode->i_sb->u.aclfs_sb.s_ninodes) {
		printk("Bad inode number on dev 0x%04x: %d is out of range\n",
			inode->i_dev, ino);
		return;
	}
	block = 2 + inode->i_sb->u.aclfs_sb.s_imap_blocks +
		    inode->i_sb->u.aclfs_sb.s_zmap_blocks +
		    (ino-1)/ACLFS_INODES_PER_BLOCK;
	if (!(bh=bread(inode->i_dev,block, BLOCK_SIZE))) {
		printk("Major problem: unable to read inode from dev 0x%04x\n",
			inode->i_dev);
		return;
	}
	raw_inode = ((struct aclfs_inode *) bh->b_data) +
		    (ino-1)%ACLFS_INODES_PER_BLOCK;
	inode->i_mode = raw_inode->i_mode;
	inode->i_uid = raw_inode->i_uid;
	inode->i_gid = raw_inode->i_gid;
	inode->i_nlink = raw_inode->i_nlinks;
	inode->i_size = raw_inode->i_size;
	inode->i_mtime = inode->i_atime = inode->i_ctime = raw_inode->i_time;
	inode->i_blocks = inode->i_blksize = 0;

	/* object entries */
	ACL(x,ugid) = 0;
	ACL(x,mode) = ACL_USER_OBJ | ACL_VALID;
	ACL(x,mode) |= ((inode->i_mode & 448) >> 1); 
	x++;
  
	ACL(x,ugid) = 0;
	ACL(x,mode) = ACL_GROUP_OBJ | ACL_VALID;
	ACL(x,mode) |= ((inode->i_mode & 56) << 2); 
	x++;

	ACL(x,ugid) = 0;
	ACL(x,mode) = ACL_OTHER_OBJ | ACL_VALID;
	ACL(x,mode) |= ((inode->i_mode & 7) << 5); 
        x++;

	/* copy over the acl information stored in the inode */
	for (y = 0; x < ACLFS_ONDISK_ACLS && raw_inode->acl_list[y].mode; x++, y++)
	{
		ACL(x,ugid) = raw_inode->acl_list[y].ugid;
		ACL(x,mode) = raw_inode->acl_list[y].mode;
	}



	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		inode->i_rdev = raw_inode->i_zone[0];
	else for (block = 0; block < 9; block++)
		inode->u.aclfs_i.i_data[block] = raw_inode->i_zone[block];
	brelse(bh);
	if (S_ISREG(inode->i_mode))
		inode->i_op = &aclfs_file_inode_operations;
	else if (S_ISDIR(inode->i_mode))
		inode->i_op = &aclfs_dir_inode_operations;
	else if (S_ISLNK(inode->i_mode))
		inode->i_op = &aclfs_symlink_inode_operations;
	else if (S_ISCHR(inode->i_mode))
		inode->i_op = &aclfs_chrdev_inode_operations;
	else if (S_ISBLK(inode->i_mode))
		inode->i_op = &aclfs_blkdev_inode_operations;
	else if (S_ISFIFO(inode->i_mode)) {
		inode->i_op = &aclfs_fifo_inode_operations;
		inode->i_pipe = 1;
		PIPE_BASE(*inode) = NULL;
		PIPE_HEAD(*inode) = PIPE_TAIL(*inode) = 0;
		PIPE_READ_WAIT(*inode) = PIPE_WRITE_WAIT(*inode) = NULL;
		PIPE_READERS(*inode) = PIPE_WRITERS(*inode) = 0;
	}
}

void aclfs_write_inode(struct inode * inode)
{
	struct buffer_head * bh;
	struct aclfs_inode * raw_inode;
	int ino, block, x, y;

#ifdef VERBOSE_INODE_OPS
        printk("write_inode() -- writing inode\n");
#endif

	ino = inode->i_ino;
	if (!ino || ino >= inode->i_sb->u.aclfs_sb.s_ninodes) {
		printk("Bad inode number on dev 0x%04x: %d is out of range\n",
			inode->i_dev, ino);
		inode->i_dirt = 0;
		return;
	}
	block = 2 + inode->i_sb->u.aclfs_sb.s_imap_blocks + inode->i_sb->u.aclfs_sb.s_zmap_blocks +
		(ino-1)/ACLFS_INODES_PER_BLOCK;
	if (!(bh=bread(inode->i_dev, block, BLOCK_SIZE)))
		panic("unable to read i-node block");
	raw_inode = ((struct aclfs_inode *)bh->b_data) +
		(ino-1)%ACLFS_INODES_PER_BLOCK;
	raw_inode->i_mode = inode->i_mode;
	raw_inode->i_uid = inode->i_uid;
	raw_inode->i_gid = inode->i_gid;
	raw_inode->i_nlinks = inode->i_nlink;
	raw_inode->i_size = inode->i_size;
	raw_inode->i_time = inode->i_mtime;

	/* copy acl over */
	for(y = 0, x = 0; x < MAX_ACLD && y < ACLFS_ONDISK_ACLS;  x++)
	{
		/* only copy non-object entries into raw inode */
		if (! (ACL(x,mode) & 
			(ACL_USER_OBJ | ACL_GROUP_OBJ | ACL_OTHER_OBJ)))
		{
			raw_inode->acl_list[y].ugid = ACL(x,ugid);
			raw_inode->acl_list[y].mode = ACL(x,mode);
			y++;
		}

		else switch( ACL(x,mode) & 
			(ACL_USER_OBJ | ACL_GROUP_OBJ | ACL_OTHER_OBJ))
		{
			case ACL_USER_OBJ:
				
				raw_inode->i_mode |= (ACL(x,mode) & 
				 	(ACL_READ|ACL_WRITE|ACL_EXECUTE)) << 1;
				break;

			case ACL_GROUP_OBJ:
				
				raw_inode->i_mode |= (ACL(x,mode) & 
				 	(ACL_READ|ACL_WRITE|ACL_EXECUTE)) >> 2;
				break;

			case ACL_OTHER_OBJ:
				
				 raw_inode->i_mode |= (ACL(x,mode) & 
				 	(ACL_READ|ACL_WRITE|ACL_EXECUTE)) >> 5;
				break;

			default:
				printk("write_inode: bad object specifier\n");
				break;

		}
	}
	


	if (S_ISCHR(inode->i_mode) || S_ISBLK(inode->i_mode))
		raw_inode->i_zone[0] = inode->i_rdev;
	else for (block = 0; block < 9; block++)
		raw_inode->i_zone[block] = inode->u.aclfs_i.i_data[block];
	inode->i_dirt=0;
	bh->b_dirt=1;
	brelse(bh);
}


/* this is a kludge until the mode and ACL objects merge */
int aclfs_notify_change(struct inode * inode)
{
	int x;


#ifdef VERBOSE_INODE_OPS
        printk("notify_change() -- writing change\n");
#endif
	
	for (x = 0; x < MAX_ACLD ; x++)
	{
		if (ACL(x,mode) & ACL_VALID)
		{
			if ((ACL(x,mode) & ACL_USER_OBJ) == ACL_USER_OBJ)
			{
				ACL(x,mode) &= (255 ^ 
					(ACL_READ | ACL_WRITE | ACL_EXECUTE));
				ACL(x,mode) |= ((inode->i_mode & 448) >> 1);
			}

			else if ((ACL(x,mode) & ACL_GROUP_OBJ) == ACL_GROUP_OBJ)
			{
				ACL(x,mode) &= (255 ^  
					(ACL_READ | ACL_WRITE | ACL_EXECUTE));
				ACL(x,mode) |= ((inode->i_mode & 56) << 2);
			}
			
			else if ((ACL(x,mode) & ACL_OTHER_OBJ) == ACL_OTHER_OBJ)
			{
				ACL(x,mode) &= (255 ^  
					(ACL_READ | ACL_WRITE | ACL_EXECUTE));
				ACL(x,mode) |= ((inode->i_mode & 7) << 5);
			}
		}
	}
	return 0;
}

