#ifndef _LINUX_ACLFS_FS_H
#define _LINUX_ACLFS_FS_H

#include <linux/acl.h>

/*
 * The aclfs filesystem constants/structures
 */

/*
 * Thanks to Kees J Bot for sending me the definitions of the new
 * aclfs filesystem (aka V2) with bigger inodes and 32-bit block
 * pointers. It's not actually implemented yet, but I'll look into
 * it.
 */

#define ACLFS_NAME_LEN 14
#define ACLFS_ROOT_INO 1

#define ACLFS_I_MAP_SLOTS	8
#define ACLFS_Z_MAP_SLOTS	8
#define ACLFS_SUPER_MAGIC	0x137E
#define ACLFS_ONDISK_ACLS	MAX_ACLD - 3	/* for object entries */

#define ACLFS_INODES_PER_BLOCK ((BLOCK_SIZE)/(sizeof (struct aclfs_inode)))
#define ACLFS_DIR_ENTRIES_PER_BLOCK ((BLOCK_SIZE)/(sizeof (struct aclfs_dir_entry)))

#define IS_ACLFS_INODE(inode) (((inode)->i_sb) && ((inode)->i_sb->s_magic == ACLFS_SUPER_MAGIC)) 

#define ACL(x,y) inode->u.aclfs_i.acl_list[x].y 

struct aclfs_inode {
	unsigned short i_mode;
	unsigned short i_uid;
	unsigned long i_size;
	unsigned long i_time;
	unsigned char i_gid;
	unsigned char i_nlinks;
	acl_s acl_list[ACLFS_ONDISK_ACLS];
	unsigned short i_zone[9];
};

/*
 * aclfs super-block data on disk
 */
struct aclfs_super_block {
	unsigned short s_ninodes;
	unsigned short s_nzones;
	unsigned short s_imap_blocks;
	unsigned short s_zmap_blocks;
	unsigned short s_firstdatazone;
	unsigned short s_log_zone_size;
	unsigned long s_max_size;
	unsigned short s_magic;
};

struct aclfs_dir_entry {
	unsigned short inode;
	char name[ACLFS_NAME_LEN];
};

extern int aclfs_lookup(struct inode * dir,const char * name, int len,
	struct inode ** result);
extern int aclfs_create(struct inode * dir,const char * name, int len, int mode,
	struct inode ** result);
extern int aclfs_mkdir(struct inode * dir, const char * name, int len, int mode);
extern int aclfs_rmdir(struct inode * dir, const char * name, int len);
extern int aclfs_unlink(struct inode * dir, const char * name, int len);
extern int aclfs_symlink(struct inode * inode, const char * name, int len,
	const char * symname);
extern int aclfs_link(struct inode * oldinode, struct inode * dir, const char * name, int len);
extern int aclfs_mknod(struct inode * dir, const char * name, int len, int mode, int rdev);
extern int aclfs_rename(struct inode * old_dir, const char * old_name, int old_len,
	struct inode * new_dir, const char * new_name, int new_len);
extern struct inode * aclfs_new_inode(const struct inode * dir);
extern void aclfs_free_inode(struct inode * inode);
extern unsigned long aclfs_count_free_inodes(struct super_block *sb);
extern int aclfs_new_block(struct super_block * sb);
extern void aclfs_free_block(struct super_block * sb, int block);
extern unsigned long aclfs_count_free_blocks(struct super_block *sb);

extern int aclfs_bmap(struct inode *,int);

extern struct buffer_head * aclfs_getblk(struct inode *, int, int);
extern struct buffer_head * aclfs_bread(struct inode *, int, int);

extern void aclfs_truncate(struct inode *);
extern void aclfs_put_super(struct super_block *);
extern struct super_block *aclfs_read_super(struct super_block *,void *);
extern void aclfs_read_inode(struct inode *);
extern void aclfs_write_inode(struct inode *);
extern void aclfs_put_inode(struct inode *);
extern void aclfs_statfs(struct super_block *, struct statfs *);
extern int aclfs_notify_change(struct inode *);

extern struct inode_operations aclfs_file_inode_operations;
extern struct inode_operations aclfs_dir_inode_operations;
extern struct inode_operations aclfs_symlink_inode_operations;
extern struct inode_operations aclfs_chrdev_inode_operations;
extern struct inode_operations aclfs_blkdev_inode_operations;
extern struct inode_operations aclfs_fifo_inode_operations;

extern struct file_operations aclfs_file_operations;
extern struct file_operations aclfs_dir_operations;

#endif
